//----------------------------------------------------------------------------
//
// File:        disk.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description: A simple program to display the contents of a TI-Disk image
//              created by AnaDisk.  It also supports extracting files from
//              the disk and optionally converting them to 'standard' PC type
//              files.
//
// Copyright (c) 1998-2002 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "diskio.hpp"
#include "fileio.hpp"
#include "option.hpp"

DBG_REGISTER ( __FILE__ );

void ConvertFilename ( char *buffer )
{
    FUNCTION_ENTRY ( NULL, "ConvertFilename", true );

    char *end = buffer;
    while ( *end ) end++;
    while ( *--end == ' ' );
    *++end = '\0';
    if ( *buffer == '-' ) *buffer = '_';
    for ( ; *buffer; buffer++ ) {
        if ( isalnum ( *buffer )) continue;
        if ( *buffer == '_' ) continue;
        if ( *buffer == '-' ) continue;
        *buffer = '_';
    }
}

USHORT GetUSHORT ( const void *_ptr )
{
    FUNCTION_ENTRY ( NULL, "GetUSHORT", true );

    const UCHAR *ptr = ( const UCHAR * ) _ptr;
    return ( USHORT ) (( ptr [0] << 8 ) | ptr [1] );
}

//
// Count the number of free sectors on the disk as indicated by the
//    allocation bitmap in the VIB
//
int FreeSectors ( const VIB *vib )
{
    FUNCTION_ENTRY ( NULL, "FreeSectors", true );

    int free = 0;

    for ( unsigned i = 0; i < SIZE ( vib->AllocationMap ); i++ ) {
        UCHAR bits = vib->AllocationMap [i];
        for ( int j = 0; j < 8; j++ ) {
            if (( bits & 1 ) == 0 ) free++;
            bits >>= 1;
        }
    }

    return free;
}

//
// Find a free sector on the disk beginning at sector 'start'
//
int FindFreeSector ( const VIB *vib, int start = 0 )
{
    FUNCTION_ENTRY ( NULL, "FindFreeSector", true );

    int index = 0;

    for ( unsigned i = 0; i < SIZE ( vib->AllocationMap ); i++ ) {
        UCHAR bits = vib->AllocationMap [i];
        for ( int j = 0; j < 8; j++ ) {
            if ((( bits & 1 ) == 0 ) && ( index >= start )) {
		return index;
	    }
            index++;
            bits >>= 1;
        }
    }

    return -1;
}

//
// Update the allocation bitmap in the VIB for the indicated sector
//
void SetSectorAllocation ( VIB *vib, int index, bool bUsed )
{
    FUNCTION_ENTRY ( NULL, "SetSectorAllocation", true );

    int i   = index / 8;
    int bit = index % 8;

    if ( bUsed == true ) {
        vib->AllocationMap [i] |= 1 << bit;
    } else {
        vib->AllocationMap [i] &= ~ ( 1 << bit );
    }
}

const sSector *FindSector ( const cDiskMedia *disk, const VIB *vib, int index )
{
    FUNCTION_ENTRY ( NULL, "FindSector", true );

    int trackSize = (( vib != NULL ) && ( vib->SectorsPerTrack != 0 )) ? vib->SectorsPerTrack : 9;

    int t = index / trackSize;
    int s = index % trackSize;
    int h = 0;

    if ( t >= disk->NumTracks ()) {
        t = 2 * disk->NumTracks () - t - 1;
        h = 1;
    }

    if ( t >= disk->NumTracks ()) {
        fprintf ( stderr, "Invalid sector index (%d)\n", index );
        exit ( -1 );
    }

    return disk->GetSector ( t, h, s );
}

int FindFileDescriptorIndex ( const cDiskMedia *disk, const VIB *vib, const char *name )
{
    FUNCTION_ENTRY ( NULL, "FindFileDescriptorIndex", true );

    USHORT *dirIndex = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    int start = ( dirIndex [0] == 0 ) ? 1 : 0;
    int len = ( strlen ( name ) < 10 ) ? strlen ( name ) : 10;

    for ( int i = start; i < 128; i++ ) {
        if ( dirIndex [i] == 0 ) break;
        int index = GetUSHORT ( &dirIndex [i] );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, index )->Data;
        if ( strnicmp ( FDR->FileName, name, len ) == 0 ) {
            return index;
        }
    }

    return -1;
}

typedef int (*QSORT_FUNC) ( const void *, const void * );

static cDiskMedia *sortDisk;
static VIB        *sortVIB;

int sortDirectoryIndex ( const void *ptr1, const void *ptr2 )
{
    sFileDescriptorRecord *fdr1 = ( sFileDescriptorRecord * ) FindSector ( sortDisk, sortVIB, GetUSHORT ( ptr1 ))->Data;
    sFileDescriptorRecord *fdr2 = ( sFileDescriptorRecord * ) FindSector ( sortDisk, sortVIB, GetUSHORT ( ptr2 ))->Data;

    return strcmp ( fdr1->FileName, fdr2->FileName );
}

int AddFileDescriptor ( cDiskMedia *disk, VIB *vib, const sFileDescriptorRecord *FDR )
{
    FUNCTION_ENTRY ( NULL, "AddFileDescriptor", true );

    USHORT *FDI = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    sortDisk = disk;
    sortVIB  = vib;

    int fdrIndex = -1;

    int start = ( FDI [0] == 0 ) ? 1 : 0;

    // Look for a free slot in the file descriptor index
    for ( int i = start; i < 127; i++ ) {
        if ( FDI [i] == 0 ) {
            // Find a place to put the FDR
            fdrIndex = FindFreeSector ( vib );
            if ( fdrIndex != -1 ) {
                // Mark the new FDR's sector as used
                SetSectorAllocation ( vib, fdrIndex, true );

                // Copy the FDR to the sector on disk
                sFileDescriptorRecord *newFDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, fdrIndex )->Data;
                memcpy ( newFDR, FDR, 256 );

                // Make sure the new name is padded with spaces
                for ( unsigned j = strlen ( newFDR->FileName ); j < sizeof ( newFDR->FileName ); j++ ) {
                    newFDR->FileName [j] = ' ';
                }

                // Add the name and resort the directory
                FDI [i] = GetUSHORT ( &fdrIndex );
                qsort ( FDI + start, i - start + 1, sizeof ( USHORT ), ( QSORT_FUNC ) sortDirectoryIndex );
            }
            break;
        }
    }

    return fdrIndex;
}

bool AddChain ( sFileDescriptorRecord *FDR, int start, int offset )
{
    FUNCTION_ENTRY ( NULL, "AddChain", true );

    CHAIN *chain = FDR->DataChain;

    for ( unsigned i = 0; i < SIZE ( FDR->DataChain ); i++ ) {
        int index = chain->start + (( int ) ( chain->start_offset & 0x0F ) << 8 );
        int total = (( int ) chain->offset << 4 ) + ( chain->start_offset >> 4 );
        if (( index == 0 ) && ( total == 0 )) {
            chain->start        = start & 0xFF;
            chain->start_offset = (( offset & 0x0F ) << 4 ) | (( start >> 8 ) & 0x0F );
            chain->offset       = ( offset >> 4 ) & 0xFF;
            return true;
        }
        chain++;
    }

    return false;
}

bool DeleteFile ( cDiskMedia *disk, VIB *vib, const char *fileName )
{
    FUNCTION_ENTRY ( NULL, "DeleteFile", true );

    int index = FindFileDescriptorIndex ( disk, vib, fileName );
    if ( index == -1 ) {
        fprintf ( stderr, "DeleteFile: File '%s' not found\n", fileName );
        return false;
    }

    USHORT *FDI = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;
    sortDisk = disk;
    sortVIB  = vib;

    int start = ( FDI [0] == 0 ) ? 1 : 0;

    int fdrIndex = GetUSHORT ( &index );
    for ( int i = start; i < 127; i++ ) {
        if ( FDI [i] == fdrIndex ) {
            memcpy ( FDI + i, FDI + i + 1, ( 128 - i - 1 ) * sizeof ( USHORT ));
            FDI [127] = 0;
            break;
        }
    }

    int last = 127;
    while (( last > start ) && ( FDI [last] == 0 )) {
        last--;
    }
    if ( last > start ) {
        qsort ( FDI + start, last - start + 1, sizeof ( USHORT ), ( QSORT_FUNC ) sortDirectoryIndex );
    }

    sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, index )->Data;

    // Release the FDR's sector
    SetSectorAllocation ( vib, index, false );

    CHAIN *chain = FDR->DataChain;
    int sectorsTotal = GetUSHORT ( &FDR->TotalSectors );

    for ( int lastOffset = 0; lastOffset < sectorsTotal; ) {
        int start  = chain->start + (( int ) ( chain->start_offset & 0x0F ) << 8 );
        int offset = (( int ) chain->offset << 4 ) + ( chain->start_offset >> 4 ) + 1;
        ASSERT (( start != 0 ) && ( offset != 0 ));
        ASSERT ( offset > lastOffset );
        // Release sectors in this chain
        for ( int i = 0; i < offset - lastOffset; i++ ) {
            SetSectorAllocation ( vib, start + i, false );
        }
        lastOffset = offset;
        chain++;
    }

    return true;
}

bool AddFile ( cDiskMedia *disk, VIB *vib, const sFileDescriptorRecord *FDR, const UCHAR *fileData )
{
    FUNCTION_ENTRY ( NULL, "AddFile", true );

    // Make sure there isn't a file already on this disk with the same name
    if ( FindFileDescriptorIndex ( disk, vib, FDR->FileName ) != -1 ) {
        DeleteFile ( disk, vib, FDR->FileName );
    }

    // First, make sure the disk has enough room for the file data
    int totalSectors = GetUSHORT ( &FDR->TotalSectors );
    if ( FreeSectors ( vib ) < totalSectors + 1 ) {
        fprintf ( stderr, "AddFile: Out of disk space - unable to add file '%s'\n", FDR->FileName );
        return false;
    }

    // Next, Try to add another filename
    int fdrIndex = AddFileDescriptor ( disk, vib, FDR );
    if ( fdrIndex == -1 ) {
        fprintf ( stderr, "AddFile: Too many files on disk - unable to add file '%s'\n", FDR->FileName );
        return false;
    }

    // Get a pointer to the new FDR so we can update the file chain
    sFileDescriptorRecord *newFDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, fdrIndex )->Data;

    // Now, add the data sectors & update the FDR
    int sectorsTotal = 0;
    int chainStart   = FindFreeSector ( vib );
    int chainLength  = 0;
    for ( int i = 0; i < totalSectors; i++ ) {
        int index = FindFreeSector ( vib );
        SetSectorAllocation ( vib, index, true );
        UCHAR *data = ( UCHAR * ) FindSector ( disk, vib, index )->Data;
        memcpy ( data, fileData, 256 );
        fileData += 256;
        if ( index != chainStart + chainLength ) {
            if ( AddChain ( newFDR, chainStart, chainLength - 1 + sectorsTotal ) == false ) {
                goto bad_chain;
            }
            sectorsTotal += chainLength;
            chainStart  = index;
            chainLength = 1;
        } else {
            chainLength++;
        }
    }

    // Add the last chain
    if ( AddChain ( newFDR, chainStart, chainLength - 1 + sectorsTotal ) == false ) {
bad_chain:
        fprintf ( stderr, "AddFile: Disk too fragmented - unable to add file '%s'\n", FDR->FileName );
        // Undo everything we just did
        DeleteFile ( disk, vib, FDR->FileName );
        // Now free up the sectors that weren't in the FDR's allocation chain
        for ( int i = 0; i < chainLength; i++ ) {
            SetSectorAllocation ( vib, chainStart + i, false );
        }
        return false;
    }

    return true;
}

void DumpSectors ( const cDiskMedia *disk, const VIB *vib, int sIndex, int sCount, bool isLast, sFileDescriptorRecord *FDR, bool convertFiles, FILE *outFile )
{
    FUNCTION_ENTRY ( NULL, "DumpSectors", true );

    bool isDisplay = (( FDR->FileStatus & INTERNAL_TYPE ) == 0 ) ? true : false;

    for ( int i = 0; i < sCount; i++ ) {

        const sSector *sector = FindSector ( disk, vib, sIndex );
        if ( sector == NULL ) {
            fprintf ( stderr, "File: %s - sector %d not found\n", FDR->FileName, sIndex );
            sIndex++;
            continue;
        }
        UCHAR *ptr = sector->Data;

        if ( convertFiles == false ) {
            fwrite ( ptr, 256, 1, outFile );
        } else {
            if ( FDR->FileStatus & PROGRAM_TYPE ) {
                int size = (( i == sCount - 1 ) && isLast && ( FDR->EOF_Offset != 0 )) ? FDR->EOF_Offset : 256;
                fwrite ( ptr, size, 1, outFile );
            } else if ( FDR->FileStatus & VARIABLE_TYPE ) {
                while ( *ptr != 0xFF ) {
                    if ( isDisplay == true ) {
                        int length = *ptr++;
                        fwrite ( ptr, length, 1, outFile );
                        fputs ( "\n", outFile );
                        ptr += length;
                    } else {
                        int length = *ptr + 1;
                        fwrite ( ptr, length, 1, outFile );
                        ptr += length;
                    }
                }
            } else {
                if ( FDR->RecordLength != 0 ) {
                    for ( int j = 0; j < FDR->RecordsPerSector; j++ ) {
                        if ( FDR->NoFixedRecords == 0 ) break;
                        FDR->NoFixedRecords--;
                        fwrite ( ptr, FDR->RecordLength, 1, outFile );
                        if ( isDisplay == true ) {
                            fputs ( "\n", outFile );
                        }
                        ptr += FDR->RecordLength;
                    }
                } else {
                    fwrite ( ptr, 256, 1, outFile );
                    if ( isDisplay == true ) {
                        fputs ( "\n", outFile );
                    }
                    ptr += 256;
                }
            }
        }
        sIndex++;
    }
}

void DumpFile ( const cDiskMedia *disk, const VIB *vib, const sFileDescriptorRecord *FDR, bool convertFiles )
{
    FUNCTION_ENTRY ( NULL, "DumpFile", true );

    char name [ 32 ];

    strcpy ( name, FDR->FileName );
    ConvertFilename ( name );

    // Decorate the name to match the file's type
    if ( convertFiles == true ) {
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
            strcat ( name, ".PROG" );
        } else {
            sprintf ( name + strlen ( name ), ".%c%c%d",
                ( FDR->FileStatus & INTERNAL_TYPE ) ? 'I' : 'D',
                ( FDR->FileStatus & VARIABLE_TYPE ) ? 'V' : 'F',
                FDR->RecordLength ? FDR->RecordLength : 256 );
        }
    }

    // Find a unique filename to use
    char buffer [ 80 ];
    FILE *outFile;
    for ( int i = 0; i < 255; i++ ) {
        sprintf ( buffer, i ? "%s.%03d" : "%s", name, i );
        outFile = fopen ( buffer, "rb" );
        if ( outFile == NULL ) break;
        fclose ( outFile );
    }

    const char *mode = (( convertFiles == true ) && (( FDR->FileStatus & ( INTERNAL_TYPE | PROGRAM_TYPE )) == 0 )) ? "wt" : "wb";

    // Open the file
    outFile = fopen ( buffer, mode );
    if ( outFile == NULL ) return;

    // Write the file header
    if ( convertFiles == false ) {
        char header [128];
        memset ( header, 0, 128 );
        memcpy ( header, FDR, 20 );
        fwrite ( header, 128, 1, outFile );
    }

    // Make a copy of the file descriptor since DumpSectors modifies NoFixedRecords
    sFileDescriptorRecord fdr;
    memcpy ( &fdr, FDR, 256 );

    int size = GetUSHORT ( &FDR->TotalSectors );
    int copied = 0;
    const CHAIN *chain = FDR->DataChain;
    do {
        int sector = chain->start + (( int ) ( chain->start_offset & 0x0F ) << 8 );
        int offset = (( int ) chain->offset << 4 ) + ( chain->start_offset >> 4 ) + 1;
        DumpSectors ( disk, vib, sector, offset - copied, offset == size, &fdr, convertFiles, outFile );
        copied = offset;
        chain++;
    } while ( copied < size );

    fclose ( outFile );
}

bool DumpFile ( const cDiskMedia *disk, const VIB *vib, const char *fileName, bool convertFiles )
{
    FUNCTION_ENTRY ( NULL, "DumpFile", true );

    int index = FindFileDescriptorIndex ( disk, vib, fileName );
    if ( index == -1 ) {
        fprintf ( stderr, "DumpFile: File '%s' not found\n", fileName );
        return false;
    }

    sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, index )->Data;
    DumpFile ( disk, vib, FDR, convertFiles );

    return true;
}

void ShowDirectory ( cDiskMedia *disk )
{
    FUNCTION_ENTRY ( NULL, "ShowDirectory", true );

    if ( disk->GetFormat () == FORMAT_UNKNOWN ) {
        printf ( "Could not read the disk format\n" );
        return;
    }

    VIB *vib = ( VIB * ) disk->GetSector ( 0, 0, 0 )->Data;

    printf ( "\n" );
    printf ( "Directory of %10.10s\n", vib->VolumeName );
    printf ( "\n" );
    printf ( "  Filename   Size Type        P\n" );
    printf ( " ==========  ==== =========== =\n" );

    USHORT *dirIndex = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    int sectorsUsed = 0;

    for ( int i = 0; i < 128; i++ ) {

        if ( dirIndex [i] == 0 ) break;

        int index = GetUSHORT ( &dirIndex [i] );

        const sSector *sector = FindSector ( disk, vib, index );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) sector->Data;

        printf ( "  %10.10s", FDR->FileName );
        int size = GetUSHORT ( &FDR->TotalSectors ) + 1;
        sectorsUsed += size;
        printf ( " %4d", size );
        if ( FDR->FileStatus & PROGRAM_TYPE ) {
            printf ( " PROGRAM    " );
        } else {
            printf ( " %s/%s %3d", ( FDR->FileStatus & INTERNAL_TYPE ) ? "INT" : "DIS",
                                   ( FDR->FileStatus & VARIABLE_TYPE ) ? "VAR" : "FIX",
                                   FDR->RecordLength ? FDR->RecordLength : 256 );
        }
        if ( FDR->FileStatus & WRITE_PROTECTED_TYPE ) {
            printf ( " Y" );
        }
        printf ( "\n" );
    }

    int totalSectors   = GetUSHORT ( &vib->FormattedSectors ) - 2;
    int totalAvailable = FreeSectors ( vib );

    printf ( "\n" );
    printf ( "  Available: %4d  Used: %4d\n", totalAvailable, sectorsUsed );
    printf ( "      Total: %4d   Bad: %4d\n", totalSectors, totalSectors - sectorsUsed - totalAvailable );
    printf ( "\n" );
}

bool FileExists ( const char *filename )
{
    FUNCTION_ENTRY ( NULL, "FileExists", true );

    FILE *file = fopen ( filename, "rb" );
    if ( file == NULL ) return false;
    fclose ( file );
    return true;
}

void PrintUsage ()
{
    FUNCTION_ENTRY ( NULL, "PrintUsage", true );

    fprintf ( stdout, "Usage: disk [options] file\n" );
    fprintf ( stdout, "\n" );
}

void PrintDiskInfo ( const cDiskMedia &disk, bool showLayout )
{
    FUNCTION_ENTRY ( NULL, "PrintDiskInfo", true );

    eDiskDensity density = DENSITY_UNKNOWN;
    eDiskDensity lastDensity = DENSITY_UNKNOWN;

    printf ( "\n" );

    for ( int h = 0; h < disk.NumSides (); h++ ) {
        for ( int t = 0; t < disk.NumTracks (); t++ ) {
            const sTrack *track = disk.GetTrack ( t, h );
            if ( track == NULL ) continue;
            if ( density == DENSITY_UNKNOWN ) {
                density = track->Density;
            } else if (( density != track->Density ) && ( track->Density != DENSITY_UNKNOWN )) {
                density = DENSITY_MIXED;
            }
            if ( showLayout == true ) {
                printf ( "Track: %2d  Side: %d - ", t, h );
                if ( track->Density == DENSITY_UNKNOWN ) {
                    printf ( "is not formatted\n" );
                    continue;
                }
                if ( lastDensity != track->Density ) {
                    if ( lastDensity != DENSITY_UNKNOWN ) {
                        printf ( "density changed -" );
                    }
                    lastDensity = track->Density;
                }
                int  noSectors = 0;
                for ( int s = 0; s < MAX_SECTORS; s++ ) {
                    if ( track->Sector [s].Data == NULL ) continue;
                    printf ( " %d", track->Sector [s].LogicalSector );
                    noSectors++;
                }
                printf ( "\n" );
                for ( int s = 0; s < MAX_SECTORS; s++ ) {
                    if ( track->Sector [s].Data == NULL ) continue;
                    if ( track->Sector [s].Data [-1] != 0xFB ) {
                        printf ( "  contains deleted sector %d\n", track->Sector [s].LogicalSector );
                    }
                    if ( track->Sector [s].LogicalCylinder != t ) {
                        printf ( "  sector %d has incorrect cylinder %d\n", track->Sector [s].LogicalSector, track->Sector [s].LogicalCylinder );
                    }
                    if ( track->Sector [s].LogicalSide != h ) {
                        printf ( "  sector %d has incorrect side %d\n", track->Sector [s].LogicalSector, track->Sector [s].LogicalSide );
                    }
                    if ( track->Sector [s].Size != 1 ) {
                        printf ( "  sector %d size is %d bytes\n", track->Sector [s].LogicalSector, 128 << track->Sector [s].Size );
                    }
                }
                if (( noSectors != 9 ) && ( noSectors != 18 )) {
                    printf ( "  contains %d sectors\n", noSectors );
                }
            }
        }
    }

    if ( showLayout == true ) {
        printf ( "\n" );
    }

    const char *strFormat = "<Unknown>";
    switch ( disk.GetFormat ()) {
        case FORMAT_RAW_TRACK :
            strFormat = "PC99";
            break;
        case FORMAT_RAW_SECTOR :
            strFormat = "v9t9";
            break;
        case FORMAT_ANADISK :
            strFormat = "AnaDisk";
            break;
        default :
            break;
    }

    const char *txtDensity = "<Unknown>";
    switch ( density ) {
        case DENSITY_SINGLE :
            txtDensity = "Single";
            break;
        case DENSITY_DOUBLE :
            txtDensity = "Double";
            break;
        case DENSITY_MIXED :
            txtDensity = "Mix";
            break;
        default :
            break;
    }

    printf ( " Format: %s\n", strFormat );
    printf ( " Tracks: %d\n", disk.NumTracks ());
    printf ( "  Sides: %d\n", disk.NumSides ());
    printf ( "Density: %s\n", txtDensity );
    printf ( "\n" );
}

void GetFilenames ( cDiskMedia *disk, VIB *vib, char *names[] )
{
    FUNCTION_ENTRY ( NULL, "GetFilenames", true );

    USHORT *dirIndex = ( USHORT * ) FindSector ( disk, vib, 1 )->Data;

    int start = ( dirIndex [0] == 0 ) ? 1 : 0;

    for ( int i = start, x = 0; i < 128; i++ ) {
        if ( dirIndex [i] == 0 ) break;
        int index = GetUSHORT ( &dirIndex [i] );
        sFileDescriptorRecord *FDR = ( sFileDescriptorRecord * ) FindSector ( disk, vib, index )->Data;
        names [x] = new char [11];
        memcpy ( names [x], FDR->FileName, 10 );
        names [x][10] = '\0';
        for ( int j = 9; j > 0; j-- ) {
            if ( names [x][j] != ' ' ) break;
            names [x][j] = '\0';
        }
        x++;
    }
}

char *addFiles [127];
char *delFiles [127];
char *extFiles [127];

bool ParseFileName ( const char *arg, void *base )
{
    FUNCTION_ENTRY ( NULL, "ParseFileName", true );

    const char *ptr = strchr ( arg, '=' );

    // Handle special case of extract all
    if ( ptr == NULL ) {
        fprintf ( stderr, "A filename needs to be specified: '%s'\n", arg );
        return false;
    }

    char **list = ( char ** ) base;

    for ( int i = 0; i < 127; i++ ) {
        if ( list [i] == NULL ) {
            list [i] = strdup ( ptr + 1 );
            return true;
        }
    }

    fprintf ( stderr, "Too many files specified\n" );

    return false;
}

int main ( int argc, char *argv[] )
{
    FUNCTION_ENTRY ( NULL, "main", true );

    bool dumpFiles    = false;
    bool convertFiles = false;
    bool verboseMode  = false;
    bool showLayout   = false;
    eDiskFormat outputFormat = FORMAT_UNKNOWN;

    sOption optList [] = {
        { 'a', "add=*<filename>",     OPT_NONE,                      true,              addFiles,      ParseFileName,  "Add <filename> to the disk image" },
        { 'c', "convert",             OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &convertFiles, NULL,           "Convert extracted files to DOS files" },
        { 'd', "dump",                OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &dumpFiles,    NULL,           "Extract all files to FIAD files" },
        { 'l', "layout",              OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &showLayout,   NULL,           "Display the disk sector layout" },
        {  0,  "output=PC99",         OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_RAW_TRACK,  &outputFormat, NULL,           "Convert disk to PC99 format" },
        {  0,  "output=v9t9",         OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_RAW_SECTOR, &outputFormat, NULL,           "Convert disk to v9t9 DOAD format" },
        {  0,  "output=anadisk",      OPT_VALUE_SET | OPT_SIZE_INT,  FORMAT_ANADISK,    &outputFormat, NULL,           "Convert disk to AnaDisk format /w headers" },
        { 'r', "remove=*<filename>",  OPT_NONE,                      true,              delFiles,      ParseFileName,  "Remove <filename> from the disk image" },
        { 'v', "verbose",             OPT_VALUE_SET | OPT_SIZE_BOOL, true,              &verboseMode,  NULL,           "Display information about the disk image" },
        { 'e', "extract*=<filename>", OPT_NONE,                      true,              extFiles,      ParseFileName,  "Extract <filename> to v9t9 FIAD file" }
    };

    if ( argc == 1 ) {
        PrintHelp ( SIZE ( optList ), optList );
        return 0;
    }

    printf ( "TI-99/4A Diskette Viewer\n" );

    const char *validName = NULL;

    int index = 1;
    while ( index < argc ) {

        index = ParseArgs ( index, argc, argv, SIZE ( optList ), optList );

        if ( index < argc ) {
            const char *fileName = LocateFile ( "disks", argv [index++] );
            if ( fileName == NULL ) {
                fprintf ( stderr, "Unable to open disk image file '%s'\n", argv [index-1] );
                return -1;
            }
            if ( validName != NULL ) {
                fprintf ( stderr, "Only one disk image file can be specified\n" );
                return -1;
            }
            validName = fileName;
        }
    }

    if ( showLayout == true ) verboseMode = true;

    if ( validName == NULL ) {
        fprintf ( stderr, "No disk image file specified\n" );
        return -1;
    }

    cDiskMedia disk ( validName );

    if ( disk.GetFormat () == FORMAT_UNKNOWN ) {
        fprintf ( stderr, "File '%s' contains an unrecognized disk format\n", disk.GetName ());
        return -1;
    }

    if ( verboseMode == true ) {
        PrintDiskInfo ( disk, showLayout );
    }

    bool diskChanged = false;

    VIB *vib = ( VIB * ) disk.GetSector ( 0, 0, 0 )->Data;

    // Add new files here
    if ( addFiles [0] != NULL ) {
        if ( verboseMode == true ) fprintf ( stdout, "\n" );
        if ( verboseMode == true ) fprintf ( stdout, "Adding files:\n" );
        for ( int i = 0; addFiles [i]; i++ ) {
            if ( verboseMode == true ) fprintf ( stdout, "  %s\n", addFiles [i] );
            // TBD - add the files
        }
    }

    // Remove files
    if ( delFiles [0] != NULL ) {
        if ( verboseMode == true ) fprintf ( stdout, "\n" );
        if ( verboseMode == true ) fprintf ( stdout, "Removing files:\n" );
        for ( int i = 0; delFiles [i]; i++ ) {
            if ( verboseMode == true ) fprintf ( stdout, "  %s\n", delFiles [i] );
            diskChanged |= DeleteFile ( &disk, vib, delFiles [i] );
        }
    }

    // Extract files
    if ( dumpFiles == true ) {
        // Ignore any files specified with '--extract=' and dump them all
        GetFilenames ( &disk, vib, extFiles );
    }

    if ( extFiles [0] != NULL ) {
        if ( verboseMode == true ) fprintf ( stdout, "\n" );
        if ( verboseMode == true ) fprintf ( stdout, "Extracting files:\n" );
        for ( int i = 0; extFiles [i]; i++ ) {
            if ( verboseMode == true ) fprintf ( stdout, "  %s\n", extFiles [i] );
            DumpFile ( &disk, vib, extFiles [i], convertFiles );
        }
    }

    ShowDirectory ( &disk );

    if (( outputFormat != FORMAT_UNKNOWN ) || ( diskChanged == true )) {
        if ( disk.SaveFile ( outputFormat, true ) == false ) {
            fprintf ( stderr, "Unable to write to file '%s'\n", disk.GetName ());
            return -1;
        }
    }

    return 0;
}
